#
# (C) Copyright IBM Corp. 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import click
import tempfile
import inquirer
import yaml 
import os.path as path

from ibm_cloud_sdk_core import ApiException
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_platform_services import GlobalSearchV2, GlobalTaggingV1
from ibm_vpc import VpcV1

from inquirer import errors

def get_option_from_list(msg, choices, default=None, choice_key='name', do_nothing=None):
    if(len(choices) == 0):
        error_msg = f"There no option for {msg}"
        print(error_msg)
        raise Exception(error_msg)

    if(len(choices) == 1 and not do_nothing):
        return(choices[0])

    choices_keys = [choice[choice_key] for choice in choices]
    if do_nothing:
        choices_keys.insert(0, do_nothing)

    questions = [
            inquirer.List('answer',
                message=msg,
                choices=choices_keys,
                default=default,
            ),]
    answers = inquirer.prompt(questions)
    
    # now find the object by name in the list
    if answers['answer'] == do_nothing:
        return do_nothing
    else:
        return next((x for x in choices if x[choice_key] == answers['answer']), None)

def validate_not_empty(answers, current):
    if not current:
        raise errors.ValidationError('', reason=f"Key name can't be empty")
    return True

def validate_exists(answers, current):
    if not current or not os.path.exists(os.path.abspath(os.path.expanduser(current))):
        raise errors.ValidationError('', reason=f"File {current} doesn't exist")
    return True

def register_ssh_key(ibm_vpc_client, resource_group_id):
    questions = [
      inquirer.Text('keyname', message='Please specify a name for the new key', validate=validate_not_empty)
    ]
    answers = inquirer.prompt(questions)
    keyname = answers['keyname']

    EXISTING_CONTENTS = 'Paste existing public key contents'
    EXISTING_PATH = 'Provide path to existing public key'
    GENERATE_NEW = 'Generate new public key'

    questions = [
            inquirer.List('answer',
                message="Please choose",
                choices=[EXISTING_PATH, EXISTING_CONTENTS, GENERATE_NEW]
            )]

    answers = inquirer.prompt(questions)
    ssh_key_data = ""
    ssh_key_path = None
    if answers["answer"] == EXISTING_CONTENTS:
        print("Registering from file contents")
        ssh_key_data = input("[\033[33m?\033[0m] Please paste the contents of your public ssh key. It should start with ssh-rsa: ")
    elif answers["answer"] == EXISTING_PATH:
        print("Register in vpc existing key from path")
        questions = [
          inquirer.Text("public_key_path", message='Please paste path to your \033[92mpublic\033[0m ssh key', validate=validate_exists)
        ]
        answers = inquirer.prompt(questions)

        with open(answers["public_key_path"], 'r') as file:
            ssh_key_data = file.read()
    else:
        print("generate new keypair")
        filename = f"id.rsa.{keyname}"
        os.system(f'ssh-keygen -b 2048 -t rsa -f {filename} -q -N ""')
        print(f"Generated\n")
        print(f"private key: {os.path.abspath(filename)}")
        print(f"public key {os.path.abspath(filename)}.pub")
        with open(f"{filename}.pub", 'r') as file:
            ssh_key_data = file.read()
        ssh_key_path = os.path.abspath(filename)

    response = ibm_vpc_client.create_key(public_key=ssh_key_data, name=keyname, resource_group={"id": resource_group_id}, type='rsa')
    result = response.get_result()
    return result['name'], result['id'], ssh_key_path

def find_name_id(objects, msg, obj_id=None, obj_name=None, default=None, do_nothing=None):
    if obj_id:
        # just validating that obj exists
        obj_name = next((obj['name'] for obj in objects if obj['id'] == obj_id), None)
        if not obj_name:
            raise Exception(f'Object with specified id {obj_id} not found')
    if obj_name:
        obj_id = next((obj['id'] for obj in objects if obj['name'] == obj_name), None)

    if not obj_id and not obj_name:
        obj = get_option_from_list(msg, objects, default=default, do_nothing=do_nothing)
        if do_nothing and obj == do_nothing:
            return None, None

        obj_id = obj['id']
        obj_name = obj['name']

    return obj_name, obj_id

def convert_to_ray(data):
    result = {'provider': {}, 'node_config': {}}

    result['provider']['region'] = data['region']
    result['provider']['zone_name'] = data['zone']
    result['provider']['endpoint'] = data['endpoint']
    result['provider']['iam_api_key'] = data['iam_api_key'] #TODO:

    result['node_config']['vpc_id'] = data['vpc_id']
    result['node_config']['resource_group_id'] = data['resource_group_id'] 
    result['node_config']['security_group_id'] = data['sec_group_id']
    result['node_config']['subnet_id'] = data['subnet_id']
    result['node_config']['key_id'] = data['ssh_key_id']
    result['node_config']['image_id'] = data['image_id']
    result['node_config']['instance_profile_name'] = data['instance_profile_name']
    result['node_config']['volume_tier_name'] = data['volume_profile_name']
    if data.get('head_ip'):
        result['node_config']['head_ip'] = data['head_ip']

    result['ssh_key_path'] = data['ssh_key_path']

    return result

def convert_to_lithops(data):
    result = {'ibm_vpc': {}}

    result['ibm_vpc']['endpoint'] = data['endpoint']
    result['ibm_vpc']['vpc_id'] = data['vpc_id']
    result['ibm_vpc']['resource_group_id'] = data['resource_group_id']
    result['ibm_vpc']['security_group_id'] = data['sec_group_id']
    result['ibm_vpc']['subnet_id'] = data['subnet_id']
    result['ibm_vpc']['key_id'] = data['ssh_key_id']
    result['ibm_vpc']['image_id'] = data['image_id']
    result['ibm_vpc']['zone_name'] = data['zone'] 

    return result

def print_to_file(format, output_file, result, input_file):
    if format:
        print(f"converting results to {format} format")

    if not output_file:
        output_file = tempfile.mkstemp(suffix = '.yaml')[1]

    if format == 'ray':
        result = convert_to_ray(result)
        
        defaults_config_path = input_file or path.abspath(path.join(__file__ ,"../../etc/gen2-connector/defaults.yaml"))
        template_config, _, node_template, _ = get_template_config(input_file)

        with open(defaults_config_path) as f: #TODO: find path
            config = yaml.safe_load(f)
            config['provider'].update(result['provider'])

            default_cluster_name = template_config.get('cluster_name', 'default')
            default_min_workers = node_template.get('min_workers', '0')
            default_max_workers = node_template.get('max_workers', '0')


            question = [
              inquirer.Text('name', message="Cluster name, either leave default or type a new one", default=default_cluster_name),
              inquirer.Text('min_workers', message="Minimum number of worker nodes", default=default_min_workers),
              inquirer.Text('max_workers', message="Maximum number of worker nodes", default=default_max_workers)
            ]

            answers = inquirer.prompt(question)
            config['cluster_name'] = answers['name']
            config['auth']['ssh_private_key'] = result['ssh_key_path']
            config['max_workers'] = int(answers['max_workers'])

            if config.get('available_node_types'):
                for available_node_type in config['available_node_types']:
                    config['available_node_types'][available_node_type]['node_config'].update(result['node_config'])

                    if not result['node_config'].get('head_ip'):
                        config['available_node_types'][available_node_type]['node_config'].pop('head_ip', None)

                    config['available_node_types'][available_node_type]['min_workers'] = int(answers['min_workers'])
                    config['available_node_types'][available_node_type]['max_workers'] = int(answers['max_workers'])
            else:
                config['available_node_types'] = {'ray_head_default': {'node_config': result['node_config']}}
                config['available_node_types']['ray_head_default']['min_workers'] = int(answers['min_workers'])
                config['available_node_types']['ray_head_default']['max_workers'] = int(answers['max_workers'])

            with open(output_file, 'w') as outfile:
                yaml.dump(config,  outfile, default_flow_style=False)
    elif format == 'lithops':
        result = convert_to_lithops(result)
     
        with open(output_file, 'w') as outfile:
            yaml.dump(result,  outfile, default_flow_style=False)
    else:
        with open(output_file, 'w') as outfile:
            yaml.dump(result,  outfile, default_flow_style=False)

    print("\n\n=================================================")
    print(f"\033[92mCluster config file: {output_file}\033[0m")
    print("=================================================")
         
# currently supported only for ray
def get_template_config(input_file):
    template_config = {}
    provider_template = {}
    node_template = {}
    node_config = {}

    if input_file:
        with open(input_file) as f:
             template_config = yaml.safe_load(f)
             provider_template = template_config.get('provider', {})
             node_configs = tuple(template_config.get('available_node_types', {}).values())

             if node_configs:
                 node_template = node_configs[0]
                 node_config = node_configs[0].get('node_config', {})

    return template_config, provider_template, node_template, node_config

def find_default(template_dict, objects, name=None, id=None):
    val = None
    if name:
        key = 'name'
        val = template_dict.get(name)
    elif id:
        key='id'
        val = template_dict.get(id)

    if val:
        obj = next((obj for obj in objects if obj[key] == val), None)
        if obj:
            return obj['name']

REQUIRED_RULES = {'outbound_tcp_all': 'selected security group is missing rule permitting outbound TCP access\n', 'outbound_udp_all': 'selected security group is missing rule permitting outbound UDP access\n', 'inbound_tcp_sg': 'selected security group is missing rule permiting inbound tcp traffic inside selected security group\n', 'inbound_tcp_22': 'selected security group is missing rule permiting inbound traffic to tcp port 22 required for ssh\n', 'inbound_tcp_6379': 'selected security group is missing rule permiting inbound traffic to tcp port 6379 required for Redis\n', 'inbound_tcp_8265': 'selected security group is missing rule permiting inbound traffic to tcp port 8265 required to access Ray Dashboard\n'}

def validate_security_group(ibm_vpc_client, sg_id):
    required_rules = REQUIRED_RULES.copy()

    sg = ibm_vpc_client.get_security_group(sg_id).get_result()

    for rule in sg['rules']:
        # check inboud and outbound rules
        if rule['direction'] == 'outbound' and rule['remote'] == {'cidr_block': '0.0.0.0/0'}:
            if rule['protocol'] == 'all':
                # outbound is fine!
                required_rules.pop('outbound_tcp_all', None)
                required_rules.pop('outbound_udp_all', None)
            elif rule['protocol'] == 'tcp':
                required_rules.pop('outbound_tcp_all', None)
            elif rule['protocol'] == 'udp':
                required_rules.pop('outbound_udp_all', None)
        elif rule['direction'] == 'inbound':
            if rule['remote'] == {'cidr_block': '0.0.0.0/0'}:
                # we interested only in all or tcp protocols
                if rule['protocol'] == 'all':
                    # there a rule permitting all traffic
                    required_rules.pop('inbound_tcp_sg', None)
                    required_rules.pop('inbound_tcp_22', None)
                    required_rules.pop('inbound_tcp_6379', None)
                    required_rules.pop('inbound_tcp_8265', None)

                elif rule['protocol'] == 'tcp':
                    if rule['port_min'] == 1 and rule['port_max'] == 65535:
                        # all ports are open
                        required_rules.pop('inbound_tcp_sg', None)
                        required_rules.pop('inbound_tcp_22', None)
                        required_rules.pop('inbound_tcp_6379', None)
                        required_rules.pop('inbound_tcp_8265', None)
                    else:
                        port_min = rule['port_min']
                        port_max = rule['port_max']
                        if port_min <= 22 and port_max >= 22:
                            required_rules.pop('inbound_tcp_22', None)
                        elif port_min <= 6379 and port_max >= 6379:
                            required_rules.pop('inbound_tcp_6379', None)
                        elif port_min <= 8265 and port_max >= 8265:
                            required_rules.pop('inbound_tcp_8265', None)

            elif rule['remote'].get('id') == sg['id']:
                # validate that inboud trafic inside group available
                if rule['protocol'] == 'all' or rule['protocol'] == 'tcp':
                    required_rules.pop('inbound_tcp_sg', None)

    return required_rules

def build_security_group_rule_prototype_model(missing_rule, sg_id=None):
    direction, protocol, port = missing_rule.split('_')
    remote = {"cidr_block": "0.0.0.0/0"}

    try:
        port = int(port)
        port_min = port
        port_max = port
    except:
        port_min = 1
        port_max = 65535

        # only valid if security group already exists
        if port == 'sg':
            if not sg_id:
                return None
            remote = {'id': sg_id}

    return {
        'direction': direction,
        'ip_version': 'ipv4',
        'protocol': protocol,
        'remote': remote,
        'port_min': port_min,
        'port_max': port_max
        }

def create_security_group(ibm_vpc_client, vpc_id, rg_id):
    rules = []
    for rule in REQUIRED_RULES.keys():
        security_group_rule_prototype_model = build_security_group_rule_prototype_model(rule)
        if security_group_rule_prototype_model:
            rules.append(security_group_rule_prototype_model)

    q = [
          inquirer.Text('name', message="Please, type a name for the new security group", validate=validate_not_empty),
          inquirer.List('answer', message='Create new security group with all required rules', choices=['yes', 'no'], default='yes')
          ]

    answers = inquirer.prompt(q)
    if answers['answer'] == 'yes':
        vpc_identity_model = {'id': vpc_id}
        sg = ibm_vpc_client.create_security_group(vpc_identity_model, name=answers['name'], resource_group={"id": rg_id}, rules=rules).get_result() 
        sg_id = sg['id']

        # add rule to open tcp traffic inside security group
        security_group_rule_prototype_model = build_security_group_rule_prototype_model('inbound_tcp_sg', sg_id=sg_id)
        res = ibm_vpc_client.create_security_group_rule(sg_id, security_group_rule_prototype_model).get_result()
        return sg_id, sg['name'] 
    else:
        return None, None

def add_rules_to_security_group(ibm_vpc_client, sg_id, sg_name, missing_rules):
    add_rule_msgs = {
            'outbound_tcp_all': f'Add rule to open all outbound TCP ports in selected security group {sg_name}',
            'outbound_udp_all': f'Add rule to open all outbound UDP ports in selected security group {sg_name}',
            'inbound_tcp_sg': f'Add rule to open inbound tcp traffic inside selected security group {sg_name}',
            'inbound_tcp_22': f'Add rule to open inbound tcp port 22 required for SSH in selected security group {sg_name}',
            'inbound_tcp_6379': f'Add rule to open inbound tcp port 6379 required for Redis in selected security group {sg_name}',
            'inbound_tcp_8265': f'Add rule to open inbound tcp port 8265 required to access Ray Dashboard in selected security group {sg_name}'}

    for missing_rule in missing_rules.keys():
        q = [
                inquirer.List('answer',
                message=add_rule_msgs[missing_rule],
                choices=['yes', 'no'],
                default='yes')
            ]
        
        answers = inquirer.prompt(q)
        if answers['answer'] == 'yes':
            security_group_rule_prototype_model = build_security_group_rule_prototype_model(missing_rule, sg_id=sg_id)
            res = ibm_vpc_client.create_security_group_rule(sg_id, security_group_rule_prototype_model).get_result()
        else:
            return False
    return True


@click.command()
@click.option('--output-file', '-o', help='Output filename to save configurations')
@click.option('--input-file', '-i', help=f'Template for new configuration, default: {path.abspath(path.join(__file__ ,"../../etc/gen2-connector/defaults.yaml"))}')
@click.option('--iam-api-key', required=True, help='IAM_API_KEY')
@click.option('--region', help='region')
@click.option('--zone', help='availability zone name')
@click.option('--vpc-id', help='vpc id')
@click.option('--sec-group-id', help='security group id')
@click.option('--subnet-id', help='subnet id')
@click.option('--ssh-key-id', help='ssh key id')
@click.option('--image-id', help='image id')
@click.option('--instance-profile-name', help='instance profile name')
@click.option('--volume-profile-name', default='general-purpose', help='volume profile name')
@click.option('--head-ip', help='head node floating ip')
@click.option('--format', type=click.Choice(['lithops', 'ray']), help='if not specified will print plain text')
def builder(output_file, input_file, iam_api_key, region, zone, vpc_id, sec_group_id, subnet_id, ssh_key_id, image_id, instance_profile_name, volume_profile_name, head_ip, format):
    print(f"\n\033[92mWelcome to vpc config export helper\033[0m\n")

    template_config, provider_template, node_template, node_config = get_template_config(input_file) 

    authenticator = IAMAuthenticator(iam_api_key)
    ibm_vpc_client = VpcV1('2021-01-19', authenticator=authenticator)

    result = {}

    # find region and endpoint
    endpoint = None
    regions_objects = ibm_vpc_client.list_regions().get_result()['regions']
    if not region:
        default = find_default(provider_template, regions_objects, name='region')
        region_obj = get_option_from_list("Choose region", regions_objects, default = default)
        region = region_obj['name']
        endpoint = region_obj['endpoint']
    else:
        # just need to find endpoint
        region_obj = next((obj for obj in regions_objects if obj['name'] == region), None)
        endpoint = region_obj['endpoint']

    ibm_vpc_client.set_service_url(endpoint  + '/v1')

    result['region'] = region
    result['endpoint'] = endpoint

    # find availability zone
    zones_objects = ibm_vpc_client.list_region_zones(region).get_result()['zones']
    zone_obj = None
    if not zone:
        default = find_default(provider_template, zones_objects, name='zone_name')
        zone_obj = get_option_from_list("Choose availability zone", zones_objects, default = default)
    else:
        zone_obj = next((obj for obj in zones_objects if obj['name'] == zone), None)

    result['zone'] = zone_obj['name']

    vpc_name = ''
    vpc_obj = None
    vpc_objects = ibm_vpc_client.list_vpcs().get_result()['vpcs']
    if not vpc_id:
        default = find_default(node_config, vpc_objects, id='vpc_id')
        vpc_obj = get_option_from_list("Choose vpc", vpc_objects, default = default)
        vpc_id = vpc_obj['id']
        vpc_name = vpc_obj['name']
    else:
        for vpc_obj in vpc_objects:
            if vpc_obj['id'] == vpc_id:
                vpc_name = vpc_obj['name']
                break

        if not vpc_name:
            raise Exception(f'Vpc with specified id {vpc_id} not found')
            
    result['vpc_name'] = vpc_name
    result['vpc_id'] = vpc_id

    result['resource_group_id'] = vpc_obj['resource_group']['id']
    result['resource_group_name'] = vpc_obj['resource_group']['name']

    ssh_key_objects = ibm_vpc_client.list_keys().get_result()['keys']
    CREATE_NEW_SSH_KEY = "Register new SSH key in IBM VPC"

    default = find_default(node_config, ssh_key_objects, id='key_id')
    ssh_key_name, ssh_key_id = find_name_id(ssh_key_objects, 'Choose ssh key', obj_id=ssh_key_id, do_nothing=CREATE_NEW_SSH_KEY, default=default)

    ssh_key_path = None
    if not ssh_key_name:
        ssh_key_name, ssh_key_id, ssh_key_path  = register_ssh_key(ibm_vpc_client, result['resource_group_id'])

    if not ssh_key_path:
        questions = [
          inquirer.Text("private_key_path", message=f'Please paste path to \033[92mprivate\033[0m ssh key matching with selected public key {ssh_key_name}', validate=validate_exists, default="~/.ssh/id_rsa")
        ]
        answers = inquirer.prompt(questions)
        ssh_key_path = os.path.abspath(os.path.expanduser(answers["private_key_path"]))

    result['ssh_key_name'] = ssh_key_name
    result['ssh_key_id'] = ssh_key_id
    result['ssh_key_path'] = ssh_key_path

    while True:

        CREATE_NEW_SECURITY_GROUP = 'Create new security group'

        sec_group_objects = ibm_vpc_client.list_security_groups().get_result()['security_groups']
        default = find_default(node_config, sec_group_objects, id='security_group_id')
        sec_group_name, sec_group_id = find_name_id(sec_group_objects, "Choose security group", obj_id=sec_group_id, do_nothing=CREATE_NEW_SECURITY_GROUP, default=default)

        if not sec_group_name:
            sec_group_id, sec_group_name = create_security_group(ibm_vpc_client, vpc_id, vpc_obj['resource_group']['id'])
            if not sec_group_id:
                continue

        errors = validate_security_group(ibm_vpc_client, sec_group_id)
        if errors:
            for val in errors.values():
                print(f"\033[91m{val}\033[0m")

            SELECT = 'Select again security group'
            UPDATE_SELECTED = 'Interactively add required rules to selected security group'

            questions = [
                inquirer.List('answer',
                    message='Selected security group is missing required rules, see error above, please choose',
                    choices=[SELECT, UPDATE_SELECTED, CREATE_NEW_SECURITY_GROUP],
                    default=default,
                ),]

            answers = inquirer.prompt(questions)
            sg_done = False

            if answers['answer'] == SELECT:
                sec_group_id = None
                continue
            elif answers['answer'] == UPDATE_SELECTED:
                # add rules to selected security group
                sg_done = add_rules_to_security_group(ibm_vpc_client, sec_group_id, sec_group_name, errors)
            else:
                # create new security group with all rules
                sec_group_id, sec_group_name = create_security_group(ibm_vpc_client, vpc_id, vpc_obj['resource_group']['id'])
                if sec_group_id:
                    sg_done == True
    
            if not sg_done:
                sec_group_id = None
                continue

            # just in case, validate again the updated/created security group
            errors = validate_security_group(ibm_vpc_client, sec_group_id)
            if not errors:
                break
            else:
                print(f'Something failed during security group rules update/create, please update the required rules manually using ibmcli or web ui and try again')
                exit(1)
        else:
            break

    result['sec_group_name'] = sec_group_name
    result['sec_group_id'] = sec_group_id

    floating_ips = ibm_vpc_client.list_floating_ips().get_result()['floating_ips']
    if head_ip:
        for ip in floating_ips:
            if ip['address'] == head_ip:
                if ip.get('target'):
                    raise Exception(f"Specified head ip {head_ip} occupied, please choose another or let ray create a new one")
                else:
                    result['head_ip'] = head_ip
                    break
    else:
        free_floating_ips = [x for x in floating_ips if not x.get('target')]
        if free_floating_ips:
            ALLOCATE_NEW_FLOATING_IP = 'Allocate new floating ip'
            head_ip_obj = get_option_from_list("Choose head ip", free_floating_ips, choice_key='address', do_nothing=ALLOCATE_NEW_FLOATING_IP)
            if head_ip_obj and (head_ip_obj != ALLOCATE_NEW_FLOATING_IP):
                result['head_ip'] = head_ip_obj['address']

    subnet_objects = ibm_vpc_client.list_subnets().get_result()['subnets']
    default = find_default(node_config, subnet_objects, id='subnet_id')
    subnet_name, subnet_id = find_name_id(subnet_objects, "Choose subnet", obj_id=subnet_id, default=default)
    result['subnet_name'] = subnet_name
    result['subnet_id'] = subnet_id

    image_objects = ibm_vpc_client.list_images().get_result()['images']
    default = find_default(node_config, image_objects, id='image_id') or 'ibm-ubuntu-20-04-minimal-amd64-2'

    image_name, image_id = find_name_id(image_objects, 'Please choose \033[92mUbuntu\033[0m 20.04 VM image, currently only Ubuntu supported', obj_id=image_id, default=default)
    result['image_name'] = image_name
    result['image_id'] = image_id

    instance_profile_objects = ibm_vpc_client.list_instance_profiles().get_result()['profiles']
    if instance_profile_name:
        # just validate
        instance_profile_name = next((obj['name'] for obj in instance_profile_objects if obj['name'] == instance_profile_name), None)
        if not instance_profile_name:
            raise Exception(f"specified instance_profile_name {instance_profile_name} not found")
    else:
        default = find_default(node_config, instance_profile_objects, name='instance_profile_name')
        obj = get_option_from_list('Carefully choose instance profile, please refer to https://cloud.ibm.com/docs/vpc?topic=vpc-profiles', instance_profile_objects, default=default)
        instance_profile_name = obj['name']

    result['instance_profile_name'] = instance_profile_name

    result['volume_profile_name'] = volume_profile_name
    result['iam_api_key'] = iam_api_key

    print(f"vpc name: {vpc_name} id: {vpc_id}\nzone: {zone_obj['name']}\nendpoint: {endpoint}\nregion: {region}\nresource group name: {result['resource_group_name']} id: {result['resource_group_id']}\nsecurity group name: {sec_group_name} id: {sec_group_id}\nsubnet name: {subnet_name} id: {subnet_id}\nssh key name: {ssh_key_name} id {ssh_key_id}\nimage name: {image_name} id: {image_id}\n")

    print_to_file(format, output_file, result, input_file)


if __name__ == '__main__':
    builder()



#
# (C) Copyright IBM Corp. 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import inspect

import logging
import time
from uuid import uuid4

import re
import threading

from ibm_cloud_sdk_core import ApiException
from ibm_cloud_sdk_core.authenticators import IAMAuthenticator
from ibm_platform_services import GlobalSearchV2, GlobalTaggingV1
from ibm_vpc import VpcV1
from ray.autoscaler.node_provider import NodeProvider
from ray.autoscaler.tags import TAG_RAY_CLUSTER_NAME, TAG_RAY_NODE_NAME, TAG_RAY_NODE_KIND, TAG_RAY_LAUNCH_CONFIG
from ray.autoscaler._private.cli_logger import cli_logger


logger = logging.getLogger(__name__)

INSTANCE_NAME_UUID_LEN = 8
INSTANCE_NAME_MAX_LEN = 64

#TODO: move to constants
PROFILE_NAME_DEFAULT = 'cx2-2x4'
VOLUME_TIER_NAME_DEFAULT = 'general-purpose'
RAY_RECYCLABLE = 'ray-recyclable'

ALL_STATUS_TAGS = ['ray-node-status:uninitialized', 'ray-node-status:waiting-for-ssh', 'ray-node-status:setting-up', 'ray-node-status:syncing-files', 'ray-node-status:up-to-date', 'ray-node-status:update-failed']
RETRIES = 6

def _create_vpc_client(endpoint, authenticator):
    """
    Creates an IBM VPC python-sdk instance
    """
    ibm_vpc_client = VpcV1('2021-01-19', authenticator=authenticator)
    ibm_vpc_client.set_service_url(endpoint + '/v1')

    return ibm_vpc_client


class Gen2NodeProvider(NodeProvider):

    def log_in_out(func):

      def decorated_func(*args, **kwargs):
        name = func.__name__
        logger.info(f"Enter {name} from {inspect.stack()[0][3]} {inspect.stack()[1][3]}  {inspect.stack()[2][3]} with args: {args} and kwargs {kwargs}")
        try:
            result = func(*args, **kwargs)
            logger.info(f"Leave {name} from {inspect.stack()[1][3]} with result {result}")
        except:
            cli_logger.error(f"Error in {name}")
            raise
        return result
      return decorated_func

    @log_in_out
    def _init(self):
        #self.authenticator = IAMAuthenticator(iam_api_key)

        self.ibm_vpc_client = _create_vpc_client(self.endpoint, IAMAuthenticator(self.iam_api_key))
        self.global_tagging_service = GlobalTaggingV1(IAMAuthenticator(self.iam_api_key))
        self.global_search_service = GlobalSearchV2(IAMAuthenticator(self.iam_api_key))

    def __init__(self, provider_config, cluster_name):
        NodeProvider.__init__(self, provider_config, cluster_name)

        self.endpoint = self.provider_config["endpoint"]
        self.iam_api_key = self.provider_config["iam_api_key"]

        self._init()

        self.lock = threading.RLock()
        # Cache of node objects from the last nodes() call
        self.cached_nodes = {}
        self.pending_nodes = {}
        self.cached_by_query = {}

        self.cache_stopped_nodes = provider_config.get("cache_stopped_nodes", True)

    def _get_node_type(self, node):
        for tag in node['tags']:
            kv = tag.split(':')
            if kv[0] == 'ray-node-type':
                return kv[1]

    @log_in_out
    def non_terminated_nodes(self, tag_filters):
        # get all nodes tagged for cluster "query":"tags:\"ray-cluster-name:default\" AND tags:\"ray-node-type:head\""
        CLUSTER_NAME_QUERY = f"tags:\"{TAG_RAY_CLUSTER_NAME}:{self.cluster_name}\""
        query = CLUSTER_NAME_QUERY
        for k, v in sorted(tag_filters.items()):
            query += f" AND tags:\"{k}:{v}\""

        with self.lock:
            # retry search few times in case node misteriously dissapeared
            NT_RETRIES = 3
            for retry in range(NT_RETRIES):
                try:
                    logger.info(f'self.global_search_service.search with query {query}')
                    result = self.global_search_service.search(query=query, fields=['*']).get_result()
#                    logger.info(f'query result: {result}')
                except ApiException as e:
                    logger.error(f"failed to query global search service with message: {e.message}, reiniting now")
                    self.global_search_service = GlobalSearchV2(IAMAuthenticator(self.iam_api_key))
                    result = self.global_search_service.search(query=query, fields=['*']).get_result()

                items = result['items']
                nodes = []

                # update objects with floating ips. 
                for node in items:

                    # TODO: should be special case here in case of cache_stopped_nodes?
                    if node['doc']['status'] not in ['pending', 'starting', 'running']:
                        logger.info(f"{node['resource_id']} status {node['doc']['status']} not in ['pending', 'starting', 'running'], skipping")
                        continue

                    instance_id = node['resource_id']
                    try:
                        logger.info(f"getting instance {instance_id}")
                        # this one needed to filter out zombie resources
                        self.ibm_vpc_client.get_instance(instance_id)
                    except Exception as e:
                        cli_logger.warning(instance_id)
                        if e.message == 'Instance not found':
                            logger.error(f"failed to find instance {instance_id}, skipping")
                            continue
                        logger.error(f"failed to find instance {instance_id}, raising")
                        raise e

                    node_type = self._get_node_type(node)
                    if node_type == 'head':
                        nic_id = node['doc']['network_interfaces'][0]['id']
                        res = self.ibm_vpc_client.list_instance_network_interface_floating_ips(instance_id, nic_id).get_result()

                        floating_ips = res['floating_ips']
                        if len(floating_ips) == 0:
                            # not adding a head node that is missing floating ip
                            continue
                        else:
                            # currently head node should always have floating ip.
                            # in case floating ip present we want to add it
                            node['floating_ips'] = floating_ips
                    nodes.append(node)

                found_nodes = {node["resource_id"]: node for node in nodes}

                # check if previously cached node present in the list of found nodes
                logger.inf0(f'self.cached_by_query: {self.cached_by_query}')
                _prev_nodes = self.cached_by_query.get(query, {})

                logger.info(f'_prev_nodes: {_prev_nodes.keys()} vs found_nodes {found_nodes.keys()}')
                if _prev_nodes.keys() - found_nodes.keys():
                    # nodes shouldn't dissapear too frequently, if happened most likely it is a glitch in search, retry
                    logger.warning(f'{_prev_nodes.keys() - found_nodes.keys()} from perviously cached nodes {_prev_nodes} is missing in the list of found_nodes {found_nodes.keys()}, retrying search')
                    time.sleep(1)
                    continue
                else:
                    logger.info("new nodes dont miss any perviously cached nodes")
                    break

                self.cached_by_query[query] = {node["resource_id"]: node for node in nodes}

            # update cached_nodes only for whole cluster query
            if CLUSTER_NAME_QUERY == query:
                self.cached_nodes = found_nodes

            for key in found_nodes.keys() - self.cached_nodes.keys():
                self.cached_nodes[key] = found_nodes[key]
    
            node_ids = [node["resource_id"] for node in nodes]

            # workaround for issues woth tagging/searching
            for pend_id in list(self.pending_nodes.keys()):
                if pend_id in found_nodes:
                    if found_nodes[pend_id]['doc']['status'] == 'running':
                        #the instance seems to be running, lets remove it from pendings
                        self.pending_nodes.pop(pend_id, None)
                else:
                    # if instance in pendings for more than PENDING_TIMEOUT it is time to delete it
                    PENDING_TIMEOUT = 120
                    pending_time = self.pending_nodes[pend_id] - time.time()
                    logger.info(f'the instance {pend_id} is in pendings list for {pending_time}')
                    if pending_time > PENDING_TIMEOUT:
                        logger.error(f'pending timeout {PENDING_TIMEOUT} reached, deleting instance {pend_id}')
                        self._delete_node(pend_id)
                        self.pending_nodes.pop(pend_id)
                    else:
                        node_ids.append(pend_id)

        return node_ids

    @log_in_out
    def is_running(self, node_id):
        with self.lock:
            node = self._get_cached_node(node_id)
            return node['doc']['status'] == 'running'

    @log_in_out
    def is_terminated(self, node_id):
        with self.lock:
            node = self._get_cached_node(node_id)
            state = node['doc']['status']
            return state not in ["running", "starting", "pending"]

    def _tags_to_dict(self, tags_list):
        tags_dict = {}
        #logger.info(f'in _tags_to_dict with {tags_list}')
        for tag in tags_list:
            _tags_split = tag.split(':')
#            logger.info(f'in _tags_to_dict - {_tags_split}')
            tags_dict[_tags_split[0]] = _tags_split[1]

        return tags_dict

    @log_in_out
    def node_tags(self, node_id):
        with self.lock:
            node = self._get_cached_node(node_id)
            return self._tags_to_dict(node['tags'])

    def _get_hybrid_ip(self, node_id):
        node = self._get_cached_node(node_id)
        node_type = self._get_node_type(node)
        if node_type == 'head':
            fip = node.get("floating_ips")
            if fip:
                return fip[0]['address']

            node = self._get_node(node_id)
            fip = node.get("floating_ips")
            if fip:
                return fip[0]['address']
        else:
            return self.internal_ip(node_id)

    @log_in_out
    def external_ip(self, node_id):
        with self.lock:
            if self.provider_config.get('use_hybrid_ips'):
                return self._get_hybrid_ip(node_id)

            node = self._get_cached_node(node_id)

            fip = node.get("floating_ips")
            if fip:
                return fip[0]['address']
            
            node = self._get_node(node_id)
            fip = node.get("floating_ips")
            if fip:
                return fip[0]['address']

    def internal_ip(self, node_id):
        node = self._get_cached_node(node_id)

        try:
            primary_ipv4_address = node['doc']['network_interfaces'][0].get('primary_ipv4_address')
            if primary_ipv4_address is None:
                node = self._get_node(node_id)
        except Exception:
            node = self._get_node(node_id)

        return node['doc']['network_interfaces'][0].get('primary_ipv4_address')
    
    def _global_tagging_with_retries(self, resource_crn, f_name, tags=None):
        e = None
        resource_model = {'resource_id': resource_crn}
        for retry in range(RETRIES):
            logger.info(f'retry #{retry} to {f_name} {resource_crn}')
            try:
                if f_name == 'attach_tag':
                    tag_results = self.global_tagging_service.attach_tag(
                        resources=[resource_model],
                        tag_names=tags,
                        tag_type='user').get_result()
                    if not tag_results['results'][0]['is_error']:
                        return tag_results
                elif f_name == 'list_tags':
                    return self.global_tagging_service.list_tags(attached_to=resource_crn).get_result()['items']
                else:
                    #detach
                    tag_results = self.global_tagging_service.detach_tag(
                        resources=[resource_model],
                        tag_names=tags,
                        tag_type='user').get_result()
                    if not tag_results['results'][0]['is_error']:
                        return tag_results
                logger.error(f"failed to {f_name} tags {tags} for {resource_model}, {tag_results}")
            except Exception as e:
                logger.error(f"Failed to {f_name} {tags} for {resource_model}, {e}")

            if retry == RETRIES / 2:
                # try to reinit client
                self.global_tagging_service = GlobalTaggingV1(IAMAuthenticator(self.iam_api_key))

            time.sleep(10)

        # if exception occured and reinit didn't help, raise exception
        if e:
            raise e
        else:
            return False

    @log_in_out
    def set_node_tags(self, node_id, tags):
        node = self._get_node(node_id)
        with self.lock:
            resource_crn = node['crn']

            # detach all status tags
            if not self._global_tagging_with_retries(resource_crn, 'detach_tag', tags=ALL_STATUS_TAGS):
                logger.error("failed to detach old status tags")
                raise

            logger.info("===========================")
            logger.info(f"detached {ALL_STATUS_TAGS}")
            logger.info("===========================")

            time.sleep(1)

            # attach new tags
            _tags = [f"{k}:{v}" for k,v in tags.items()]
            if not self._global_tagging_with_retries(resource_crn, 'attach_tag', tags=_tags):
                cli_logger.error(f"failed to tag node {node_id} with tags {_tags}, raising error")
                raise

            logger.info("===========================")
            logger.info(f"attached {_tags}")
            logger.info("===========================")

#            # get latest tags
#            node_tags = [t['name'] for t in self._global_tagging_with_retries(node['crn'], 'list_tags')]
#            tags_dict = self._tags_to_dict(node_tags)
#
#            # remove from dict those that not overwritten
#            for old_tag_key in list(tags_dict.keys()):
#                if old_tag_key not in tags or tags.get(old_tag_key) == tags_dict[old_tag_key]:
#                    del tags_dict[old_tag_key]
#            
#            # convert to gen2 format
#            old_tags = [f"{k}:{v}" for k,v in tags_dict.items()]
#
#            # and now detach old tags. yes, it probably may break few things, because set_tags is not atomic
#            if old_tags:
#                if not self._global_tagging_with_retries(resource_crn, 'detach_tag', tags=old_tags):
#                    cli_logger.error(f"failed to untag node {node_id} with tags {old_tags}, raising error")
#                    raise
#
#                logger.info("===========================")
#                logger.info(f"detached {old_tags}")
#                logger.info("===========================")
#
            return {} 

    def _get_instance_data(self, name):
        """
        Returns the instance information
        """
        instances_data = self.ibm_vpc_client.list_instances(name=name).get_result()
        if len(instances_data['instances']) > 0:
            return instances_data['instances'][0]            
        return None

    def _create_instance(self, name, base_config):
        """
        Creates a new VM instance

        TODO: consider to use gen2 template file instead of generating dict
        """
        logger.debug("Creating new VM instance {}".format(name))

        security_group_identity_model = {'id': base_config['security_group_id']}
        subnet_identity_model = {'id': base_config['subnet_id']}
        primary_network_interface = {
            'name': 'eth0',
            'subnet': subnet_identity_model,
            'security_groups': [security_group_identity_model]
        }

        boot_volume_profile = {
            'capacity': 100,
            'name': '{}-boot'.format(name),
            'profile': {'name': base_config.get('volume_tier_name', VOLUME_TIER_NAME_DEFAULT)}}

        boot_volume_attachment = {
            'delete_volume_on_instance_delete': True,
            'volume': boot_volume_profile
        }

        key_identity_model = {'id': base_config['key_id']}
        profile_name = base_config.get('instance_profile_name', PROFILE_NAME_DEFAULT)

        instance_prototype = {}
        instance_prototype['name'] = name
        instance_prototype['keys'] = [key_identity_model]
        instance_prototype['profile'] = {'name': profile_name}
        instance_prototype['resource_group'] = {'id': base_config['resource_group_id']}
        instance_prototype['vpc'] = {'id': base_config['vpc_id']}
        instance_prototype['image'] = {'id': base_config['image_id']}
        instance_prototype['zone'] = {'name': self.provider_config['zone_name']}
        instance_prototype['boot_volume_attachment'] = boot_volume_attachment
        instance_prototype['primary_network_interface'] = primary_network_interface

        try:
            resp = self.ibm_vpc_client.create_instance(instance_prototype)
        except ApiException as e:
            if e.code == 400 and 'already exists' in e.message:
                return self._get_instance_data(name)
            elif e.code == 400 and 'over quota' in e.message:
                cli_logger.error("Create VM instance {} failed due to quota limit"
                             .format(name))
            else:
                cli_logger.error("Create VM instance {} failed with status code {}"
                             .format(name, str(e.code)))
            raise e

        logger.info("VM instance {} created successfully ".format(name))

        return resp.result

    def _create_floating_ip(self, base_config):
        """
        Creates or attaches floating IP address
        """
        if base_config.get('head_ip'):
            for ip in self.ibm_vpc_client.list_floating_ips().get_result()['floating_ips']:
                if ip['address'] == base_config['head_ip']:
                    return ip

        floating_ip_name = '{}-{}'.format(RAY_RECYCLABLE, uuid4().hex[:4]) 
            
        logger.debug('Creating floating IP {}'.format(floating_ip_name))
        floating_ip_prototype = {}
        floating_ip_prototype['name'] = floating_ip_name
        floating_ip_prototype['zone'] = {'name': self.provider_config['zone_name']}
        floating_ip_prototype['resource_group'] = {'id': base_config['resource_group_id']}
        response = self.ibm_vpc_client.create_floating_ip(floating_ip_prototype)
        floating_ip_data = response.result

        return floating_ip_data

    def _attach_floating_ip(self, instance, fip_data):
        fip = fip_data['address']
        fip_id = fip_data['id']

        logger.debug('Attaching floating IP {} to VM instance {}'.format(fip, instance['id']))

        # we need to check if floating ip is not attached already. if not, attach it to instance
        instance_primary_ni = instance['primary_network_interface']

        if instance_primary_ni['primary_ipv4_address'] and instance_primary_ni['id'] == fip_id:
            # floating ip already attached. do nothing
            logger.debug('Floating IP {} already attached to eth0'.format(fip))
        else:
            self.ibm_vpc_client.add_instance_network_interface_floating_ip(
                instance['id'], instance['network_interfaces'][0]['id'], fip_id)        

    def _stopped_nodes(self, tags):
        # get all nodes tagged for cluster "query":"tags:\"ray-cluster-name:default\" AND tags:\"ray-node-type:head\""
        query = f"tags:\"{TAG_RAY_CLUSTER_NAME}:{self.cluster_name}\""
        query += f" AND tags:\"{TAG_RAY_NODE_KIND}:{tags[TAG_RAY_NODE_KIND]}\""
        query += f" AND tags:\"{TAG_RAY_LAUNCH_CONFIG}:{tags[TAG_RAY_LAUNCH_CONFIG]}\""

        with self.lock:
            try:
                result = self.global_search_service.search(query=query, fields=['*']).get_result()
            except ApiException as e:
                cli_logger.warning(f"failed to query global search service with message: {e.message}, reiniting now")
                self.global_search_service = GlobalSearchV2(IAMAuthenticator(self.iam_api_key))
                result = self.global_search_service.search(query=query, fields=['*']).get_result()

            items = result['items']
            nodes = []

            # filter instances by state (stopped,stopping)
            for node in items:
                instance_id = node['resource_id']
                try:
                    instance = self.ibm_vpc_client.get_instance(instance_id)
                    state = node['doc']['status']
                    if state in ["stopped", "stopping"]:
                        nodes.append(node)
                except Exception as e:
                    cli_logger.warning(instance_id)
                    if e.message == 'Instance not found':
                       continue
                    raise e
            return nodes

    def _create_node(self, base_config, tags):
        name_tag = tags[TAG_RAY_NODE_NAME]
        assert (len(name_tag) <=
                (INSTANCE_NAME_MAX_LEN - INSTANCE_NAME_UUID_LEN - 1)) and re.match("^[a-z0-9-:-]*$", name_tag), (
                    name_tag, len(name_tag))

        name = "{name_tag}-{uuid}".format(
                                name_tag=name_tag,
                                uuid=uuid4().hex[:INSTANCE_NAME_UUID_LEN])
        logger.info(f"self._create_instance with {name}")
        instance = self._create_instance(name, base_config)

        tags[TAG_RAY_CLUSTER_NAME] = self.cluster_name

        _tags = [f"{k}:{v}" for k,v in tags.items()]
        if not self._global_tagging_with_retries(instance['crn'], 'attach_tag', tags=_tags):
            # terminate the weird node that failed to be tagged
            logger.error(f"failed to tag node {instance} with tags {_tags}, terminating instance")
            self._delete_node(instance['id'])
            return {}

        # currently always creating public ip for head node
        if tags['ray-node-type'] == 'head':# or not self.provider_config.get("use_internal_ips", False):
            fip_data = self._create_floating_ip(base_config)
            self._attach_floating_ip(instance, fip_data)

        return {instance['id']: instance}

    @log_in_out
    def create_node(self, base_config, tags, count) -> None:
        with self.lock:

            # here will be implementation of reuse of stopped instances
            stopped_nodes_dict = {}
        
            # Try to reuse previously stopped nodes with compatible configs
            if self.cache_stopped_nodes:
                stopped_nodes = self._stopped_nodes(tags)
                stopped_nodes_ids = [n['resource_id'] for n in stopped_nodes]
                stopped_nodes_dict = {n['resource_id']: n for n in stopped_nodes}

                if stopped_nodes:
                    cli_logger.print(f"Reusing nodes {stopped_nodes_ids}. "
                    "To disable reuse, set `cache_stopped_nodes: False` "
                    "under `provider` in the cluster configuration.")

                for node in stopped_nodes:
                    logger.info(f"Starting instance {node['resource_id']}")
                    self.ibm_vpc_client.create_instance_action(node['resource_id'], 'start')

                    # due to issues with tagging/searching services lets keep this node in pendings for now
                    self.pending_nodes[node['resource_id']] = time.time()

                time.sleep(1)

                for node_id in stopped_nodes_ids:
                    #import pdb;pdb.set_trace()
                    self.set_node_tags(node_id, tags)

                count -= len(stopped_nodes_ids)

            created_nodes_dict = {}
            while count > 0:
                created_nodes_dict = self._create_node(base_config, tags)

                # due to issues with tagging/searching services lets keep this node in pendings for now
                for k in created_nodes_dict:
                    self.pending_nodes[k] = time.time()
                    count -= 1


            all_created_nodes = stopped_nodes_dict
            all_created_nodes.update(created_nodes_dict)
            return all_created_nodes

    def _delete_node(self, resource_id):
        logger.error(f'in _delete_node with id {resource_id}')
        try:
            node = self._get_node(resource_id)
            self.ibm_vpc_client.delete_instance(resource_id)
            floating_ips = node.get('floating_ips', [])
            for ip in floating_ips:
                if ip['name'].startswith(RAY_RECYCLABLE): # TODO: replace prefix with tags
                    self.ibm_vpc_client.delete_floating_ip(ip['id'])
        except ApiException as e:
            if e.code == 404:
                pass
            else:
                raise e

    @log_in_out
    def terminate_node(self, node_id):  # TODO: implement  terminate_nodes() as well to optimize. use instance group instead of single instance
        """
        Deletes the VM instance and the associated volume
        """
        logger.debug("Deleting VM instance {}".format(node_id))

        with self.lock:
            """
            Deletes the VM instance and the associated volume
            """
            try:
                if self.cache_stopped_nodes:
                    cli_logger.print(f"Stopping instance {node_id}. To terminate instead, set `cache_stopped_nodes: False` "
                                    "under `provider` in the cluster configuration")

                    self.ibm_vpc_client.create_instance_action(node_id, 'stop')
                else:
                    self._delete_node(node_id)
            except ApiException as e:
                if e.code == 404:
                    pass
                else:
                    raise e

    def _get_node(self, node_id):
        """Refresh and get info for this node, updating the cache."""
        self.non_terminated_nodes({})  # Side effect: updates cache

        with self.lock:
            if node_id in self.cached_nodes:
                return self.cached_nodes[node_id]

            try:
                node = self.ibm_vpc_client.get_instance(node_id).get_result()
            except Exception as e:
                logger.error(f'failed to get instance with id {node_id}')
                raise e
            try:
                node_tags = self._global_tagging_with_retries(node['crn'], 'list_tags')
                # TODO: what to do with a weird node that failed to list tags for?

            except Exception as e:
                print(f'Error listing tags for node {node}: {e}')

            return {"resource_id": node['id'], 'crn':  node['crn'], 'doc': node, 'tags': [t['name'] for t in node_tags]}

    def _get_cached_node(self, node_id):
        """Return node info from cache if possible, otherwise fetches it."""
        if node_id in self.cached_nodes:
            return self.cached_nodes[node_id]

        return self._get_node(node_id)


    @staticmethod
    def bootstrap_config(cluster_config):
        return cluster_config
